from django.db.models.fields import Field
from django.db.models.lookups import (
    Contains,
    EndsWith,
    Exact,
    IEndsWith,
    IExact,
    IRegex,
    IStartsWith,
    Regex,
    StartsWith,
)


@Field.register_lookup
class NeExact(Exact):
    lookup_name = 'neexact'


@Field.register_lookup
class NeIExact(IExact):
    lookup_name = 'neiexact'


@Field.register_lookup
class NeContains(Contains):
    lookup_name = 'necontains'


@Field.register_lookup
class NeIContains(NeContains):
    lookup_name = 'neicontains'


@Field.register_lookup
class NeStartsWith(StartsWith):
    lookup_name = 'nestartswith'


@Field.register_lookup
class NeIStartsWith(IStartsWith):
    lookup_name = 'neistartswith'


@Field.register_lookup
class NeEndsWith(EndsWith):
    lookup_name = 'neendswith'


@Field.register_lookup
class NeIEndsWith(IEndsWith):
    lookup_name = 'neiendswith'


@Field.register_lookup
class NeRegex(Regex):
    lookup_name = 'neregex'


@Field.register_lookup
class NeIRegex(IRegex):
    lookup_name = 'neiregex'



from django.db import connection
from django.test.testcases import _deferredSkip


def skipIfDBVendor(*vendors):
    return _deferredSkip(
        lambda: any(connection.vendor == vendor for vendor in vendors),
        "Database has vendor(s) %s" % ", ".join(vendors)
    )



from django.db.models.query import QuerySet as DjangoQuerySet

from .sql import Query


class QuerySet(DjangoQuerySet):
    def __init__(self, model=None, query=None, **kwargs):
        super(QuerySet, self).__init__(model=model, query=query or Query(model), **kwargs)



from django.db.models.fields import Field
from django.db.models.lookups import Regex


class AbstractExRegexLookup(Regex):
    def process_rhs(self, compiler, connection):
        rhs, params = super(AbstractExRegexLookup, self).process_rhs(compiler, connection)
        if self.rhs_is_direct_value() and params and not self.bilateral_transforms:
            for escape, operator in connection.regex_synonyms.items():
                params[0] = params[0].replace(escape, operator)
        return rhs, params


@Field.register_lookup
class ExRegexLookup(AbstractExRegexLookup):
    lookup_name = 'exregex'


@Field.register_lookup
class ExIRegexLookup(AbstractExRegexLookup):
    lookup_name = 'exiregex'


@Field.register_lookup
class NeExRegexLookup(AbstractExRegexLookup):
    lookup_name = 'neexregex'


@Field.register_lookup
class NeExIRegexLookup(AbstractExRegexLookup):
    lookup_name = 'neexiregex'



from lookup_extensions.manager import Manager

try:
    from lookup.models import (
        Article as LookupArticle,
        Game,
        Player,
        Season,
        Tag,
    )
    from or_lookups.models import (
        Article as OrLookupsArticle,
    )
    from reverse_lookup.models import (
        Choice,
        Poll,
        User,
    )
    from string_lookup.models import (
        Article as StringLookupArticle,
        Bar,
        Base,
        Child,
        Foo,
        Whiz,
    )
except ImportError:
    print("Run tests/bootstrap.py before run tests.")
    raise


def replace_managers():
    for model_cls in [
        Bar,
        Base,
        Child,
        Choice,
        Foo,
        Game,
        LookupArticle,
        OrLookupsArticle,
        Player,
        Poll,
        Season,
        StringLookupArticle,
        Tag,
        User,
        Whiz,
    ]:
        model_cls._meta.local_managers = []
        manager = Manager()
        model_cls.add_to_class('objects', manager)



from __future__ import unicode_literals

from datetime import datetime
from operator import attrgetter
from unittest import skip

from django.core.exceptions import FieldError
from django.db.models import (
    Exists,
    OuterRef,
)
from django.test import skipUnlessDBFeature

from utils import skipIfDBVendor

try:
    from lookup.models import (
        Article,
        Game,
        Player,
        Season,
        Tag,
    )
    from lookup.tests import LookupTests as DjangoLookupTests
except ImportError:
    print("Run tests/bootstrap.py before run tests.")
    raise


class LookupTests(DjangoLookupTests):
    def test_negate_lookup_int_as_str(self):
        self.assertQuerysetEqual(
            Article.objects.filter(id__neexact=str(self.a1.id)),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
            ],
        )

    @skipUnlessDBFeature('supports_date_lookup_using_string')
    def test_negate_lookup_date_as_str(self):
        self.assertQuerysetEqual(
            Article.objects.filter(pub_date__nestartswith='2005'),
            []
        )

    def test_negate_iterator(self):
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neendswith='4').iterator(),
            [
                'Article 5',
                'Article 6',
                'Article 2',
                'Article 3',
                'Article 7',
                'Article 1',
            ],
            transform=attrgetter('headline'),
        )

    def test_negate_count(self):
        self.assertEqual(Article.objects.filter(pub_date__neexact=datetime(2005, 7, 27)).count(), 4)
        self.assertEqual(Article.objects.filter(headline__nestartswith='Blah blah').count(), 7)
        self.assertEqual(Article.objects.filter(pub_date__neexact='2005-07-27 00:00:00').count(), 4)

    def test_values(self):
        self.assertSequenceEqual(
            Article.objects.filter(pub_date__neexact=datetime(2005, 7, 27)).values('id'),
            [
                {'id': self.a5.id},
                {'id': self.a6.id},
                {'id': self.a4.id},
                {'id': self.a1.id},
            ],
        )

    def test_negate_get_next_previous_by(self):
        self.assertEqual(
            repr(self.a2.get_next_by_pub_date(headline__neendswith='6')),
            '<Article: Article 3>',
        )

    def test_negate_escaping(self):
        Article.objects.create(headline='Article_ with underscore', pub_date=datetime(2005, 11, 20))

        self.assertQuerysetEqual(
            Article.objects.filter(headline__nestartswith='Article'),
            [],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__nestartswith='Article_'),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )
        Article.objects.create(headline='Article% with percent sign', pub_date=datetime(2005, 11, 21))
        self.assertQuerysetEqual(
            Article.objects.filter(headline__nestartswith='Article'),
            []
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__nestartswith='Article%'),
            [
                '<Article: Article_ with underscore>',
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ]
        )
        Article.objects.create(headline='Article with \\ backslash', pub_date=datetime(2005, 11, 22))
        self.assertQuerysetEqual(
            Article.objects.filter(headline__necontains='\\'),
            [
                '<Article: Article% with percent sign>',
                '<Article: Article_ with underscore>',
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )

    def test_negate_exclude(self):
        Article.objects.create(headline='Article_ with underscore', pub_date=datetime(2005, 11, 20))
        Article.objects.create(headline='Article% with percent sign', pub_date=datetime(2005, 11, 21))
        Article.objects.create(headline='Article with \\ backslash', pub_date=datetime(2005, 11, 22))

        self.assertQuerysetEqual(
            Article.objects.filter(headline__necontains='Article').exclude(headline__necontains='with'),
            []
        )
        self.assertQuerysetEqual(
            Article.objects.exclude(headline__nestartswith="Article_"),
            [
                '<Article: Article_ with underscore>',
            ]
        )

    def test_negate_none(self):
        self.assertQuerysetEqual(Article.objects.none().filter(headline__nestartswith='Article'), [])
        self.assertQuerysetEqual(Article.objects.filter(headline__nestartswith='Article').none(), [])

    @skip("Support NOT IN filter")
    def test_negate_in(self):
        # using __in with an empty list should return an empty query set
        self.assertQuerysetEqual(
            Article.objects.filter(id__nein=[]),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.exclude(id__nein=[]),
            [],
        )

    @skip("Support NOT IN filter")
    def test_negate_in_different_database(self):
        with self.assertRaisesMessage(
            ValueError,
            "Subqueries aren't allowed across different databases. Force the "
            "inner query to be evaluated using `list(inner_query)`."
        ):
            list(Article.objects.filter(id__nein=Article.objects.using('other').all()))

    def test_negate_regex(self):
        # Create some articles with a bit more interesting headlines for testing field lookups:
        for a in Article.objects.all():
            a.delete()
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='f')
        Article.objects.create(pub_date=now, headline='fo')
        Article.objects.create(pub_date=now, headline='foo')
        Article.objects.create(pub_date=now, headline='fooo')
        Article.objects.create(pub_date=now, headline='hey-Foo')
        Article.objects.create(pub_date=now, headline='bar')
        Article.objects.create(pub_date=now, headline='AbBa')
        Article.objects.create(pub_date=now, headline='baz')
        Article.objects.create(pub_date=now, headline='baxZ')
        # zero-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'fo*'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'fo*'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
            ],
        )
        # one-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'fo+'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: hey-Foo>'
            ],
        )
        # wildcard
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'fooo?'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: hey-Foo>',
            ],
        )
        # leading anchor
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'^b'),
            [
                '<Article: AbBa>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'^a'),
            [
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        # trailing anchor
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'z$'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'z$'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>'
            ]
        )
        # character sets
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'ba[rz]'),
            [
                '<Article: AbBa>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>'
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'ba.[RxZ]'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'ba[RxZ]'),
            [
                '<Article: AbBa>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )

        # and more articles:
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')

        # alternation
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'oo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'oo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: zoocarfaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'^foo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: barfoobaz>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',

            ],
        )

        # greedy matching
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'b.*az'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: foobar>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiregex=r'b.*ar'),
            [
                '<Article: AbBa>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: foobaz>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ]
        )

    @skipUnlessDBFeature('supports_regex_backreferencing')
    def test_negate_regex_backreferencing(self):
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neregex=r'b(.).*b\1'),
            [
                '<Article: foobar>',
                '<Article: foobaz>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )

    @skipIfDBVendor('postgresql', 'mysql')
    def test_negate_regex_null_sqlite(self):
        Season.objects.create(year=2012, gt=None)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neregex=r'^$'),
            [
                '<Season: 2012>',
            ],
        )

    @skipIfDBVendor('sqlite')
    def test_negate_regex_null_without_sqlite(self):
        Season.objects.create(year=2012, gt=None)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neregex=r'^$'),
            [],
        )

    def test_negate_regex_non_string(self):
        Season.objects.create(year=2013, gt=444)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neregex=r'^444$'),
            [],
        )

    def test_negate_regex_non_ascii(self):
        Player.objects.create(name='\u2660')
        with self.assertRaises(Player.DoesNotExist):
            Player.objects.get(name__neregex='\u2660')

    def test_negate_nonfield_lookups(self):
        msg = "Unsupported lookup 'blahblah' for CharField or join on the field not permitted."
        with self.assertRaisesMessage(FieldError, msg):
            Article.objects.filter(headline__blahblah__neexact=99)

    @skip("Support NOT IN filter")
    def test_negate_lookup_collision(self):
        # 'gt' is used as a code number for the year, e.g. 111=>2009.
        season_2009 = Season.objects.create(year=2009, gt=111)
        season_2009.games.create(home="Houston Astros", away="St. Louis Cardinals")
        season_2010 = Season.objects.create(year=2010, gt=222)
        season_2010.games.create(home="Houston Astros", away="Chicago Cubs")
        season_2010.games.create(home="Houston Astros", away="Milwaukee Brewers")
        season_2010.games.create(home="Houston Astros", away="St. Louis Cardinals")
        season_2011 = Season.objects.create(year=2011, gt=333)
        season_2011.games.create(home="Houston Astros", away="St. Louis Cardinals")
        season_2011.games.create(home="Houston Astros", away="Milwaukee Brewers")
        hunter_pence = Player.objects.create(name="Hunter Pence")
        hunter_pence.games.set(Game.objects.filter(season__year__in=[2009, 2010]))
        pudge = Player.objects.create(name="Ivan Rodriquez")
        pudge.games.set(Game.objects.filter(season__year=2009))
        pedro_feliz = Player.objects.create(name="Pedro Feliz")
        pedro_feliz.games.set(Game.objects.filter(season__year__in=[2011]))
        johnson = Player.objects.create(name="Johnson")
        johnson.games.set(Game.objects.filter(season__year__in=[2011]))

        # Games in 2010
        self.assertEqual(Game.objects.filter(season__year__exact=2010).count(), 3)
        self.assertEqual(Game.objects.filter(season__gt__exact=222).count(), 3)

        # Games in 2011
        self.assertEqual(Game.objects.filter(season__year__exact=2011).count(), 2)
        self.assertEqual(Game.objects.filter(season__gt__exact=333).count(), 2)

        # Games played in 2010 and 2011
        self.assertEqual(Game.objects.filter(season__year__in=[2010, 2011]).count(), 5)
        self.assertEqual(Game.objects.filter(season__gt__in=[222, 333]).count(), 5)

        # Players who played in 2009
        self.assertEqual(Player.objects.filter(games__season__year__exact=2009).distinct().count(), 2)
        self.assertEqual(Player.objects.filter(games__season__gt__exact=111).distinct().count(), 2)

        # Players who played in 2010
        self.assertEqual(Player.objects.filter(games__season__year__exact=2010).distinct().count(), 1)
        self.assertEqual(Player.objects.filter(games__season__gt__exact=222).distinct().count(), 1)

        # Players who played in 2011
        self.assertEqual(Player.objects.filter(games__season__year__exact=2011).distinct().count(), 2)

    @skip("Support IS NOT NULL filter")
    def test_negate_exact_none_transform(self):
        """Transforms are used for __neexact=None."""
        Season.objects.create(year=1, nulled_text_field='not null')
        self.assertFalse(Season.objects.filter(nulled_text_field__neisnull=True))
        self.assertTrue(Season.objects.filter(nulled_text_field__nulled__neisnull=True))
        self.assertTrue(Season.objects.filter(nulled_text_field__nulled__neexact=None))

    @skip("Support IS NOT NULL filter")
    def test_negate_custom_field_none_rhs(self):
        """
        __neexact=value is transformed to __neisnull=True if Field.get_prep_value()
        converts value to None.
        """
        season = Season.objects.create(year=2012, nulled_text_field=None)
        self.assertTrue(Season.objects.filter(pk=season.pk, nulled_text_field__neisnull=True))

    def test_exregex(self):
        """Copy from test_exregex."""
        # Create some articles with a bit more interesting headlines for testing field lookups:
        for a in Article.objects.all():
            a.delete()
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='f')
        Article.objects.create(pub_date=now, headline='fo')
        Article.objects.create(pub_date=now, headline='foo')
        Article.objects.create(pub_date=now, headline='fooo')
        Article.objects.create(pub_date=now, headline='hey-Foo')
        Article.objects.create(pub_date=now, headline='bar')
        Article.objects.create(pub_date=now, headline='AbBa')
        Article.objects.create(pub_date=now, headline='baz')
        Article.objects.create(pub_date=now, headline='baxZ')
        # zero-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'fo*'),
            ['<Article: f>', '<Article: fo>', '<Article: foo>', '<Article: fooo>']
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'fo*'),
            [
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ]
        )
        # one-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'fo+'),
            ['<Article: fo>', '<Article: foo>', '<Article: fooo>']
        )
        # wildcard
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'fooo?'),
            ['<Article: foo>', '<Article: fooo>']
        )
        # leading anchor
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'^b'),
            ['<Article: bar>', '<Article: baxZ>', '<Article: baz>']
        )
        self.assertQuerysetEqual(Article.objects.filter(headline__exiregex=r'^a'), ['<Article: AbBa>'])
        # trailing anchor
        self.assertQuerysetEqual(Article.objects.filter(headline__exregex=r'z$'), ['<Article: baz>'])
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'z$'),
            ['<Article: baxZ>', '<Article: baz>']
        )
        # character sets
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'ba[rz]'),
            ['<Article: bar>', '<Article: baz>']
        )
        self.assertQuerysetEqual(Article.objects.filter(headline__exregex=r'ba.[RxZ]'), ['<Article: baxZ>'])
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'ba[RxZ]'),
            ['<Article: bar>', '<Article: baxZ>', '<Article: baz>']
        )

        # and more articles:
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')

        # alternation
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'oo(f|b)'),
            [
                '<Article: barfoobaz>',
                '<Article: foobar>',
                '<Article: foobarbaz>',
                '<Article: foobaz>',
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'oo(f|b)'),
            [
                '<Article: barfoobaz>',
                '<Article: foobar>',
                '<Article: foobarbaz>',
                '<Article: foobaz>',
                '<Article: ooF>',
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'^foo(f|b)'),
            ['<Article: foobar>', '<Article: foobarbaz>', '<Article: foobaz>']
        )

        # greedy matching
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'b.*az'),
            [
                '<Article: barfoobaz>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: foobarbaz>',
                '<Article: foobaz>',
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'b.*ar'),
            [
                '<Article: bar>',
                '<Article: barfoobaz>',
                '<Article: bazbaRFOO>',
                '<Article: foobar>',
                '<Article: foobarbaz>',
            ]
        )

    @skipUnlessDBFeature('supports_exregex_backreferencing')
    def test_exregex_backreferencing(self):
        """Copy from test_regex_backreferencing."""
        # grouping and backreferences
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'b(.).*b\1'),
            ['<Article: barfoobaz>', '<Article: bazbaRFOO>', '<Article: foobarbaz>']
        )

    def test_exregex_null(self):
        """Copy from test_regex_null.

        A exregex lookup does not fail on null/None values
        """
        Season.objects.create(year=2012, gt=None)
        self.assertQuerysetEqual(Season.objects.filter(gt__exregex=r'^$'), [])

    def test_exregex_non_string(self):
        """Copy from test_regex_non_string.

        A exregex lookup does not fail on non-string fields
        """
        Season.objects.create(year=2013, gt=444)
        self.assertQuerysetEqual(Season.objects.filter(gt__exregex=r'^444$'), ['<Season: 2013>'])

    def test_exregex_non_ascii(self):
        """Copy from test_regex_non_ascii.

        A exregex lookup does not trip on non-ASCII characters.
        """
        Player.objects.create(name='\u2660')
        Player.objects.get(name__exregex='\u2660')

    def test_negate_exregex(self):
        """Copy from test_negate_regex."""
        # Create some articles with a bit more interesting headlines for testing field lookups:
        for a in Article.objects.all():
            a.delete()
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='f')
        Article.objects.create(pub_date=now, headline='fo')
        Article.objects.create(pub_date=now, headline='foo')
        Article.objects.create(pub_date=now, headline='fooo')
        Article.objects.create(pub_date=now, headline='hey-Foo')
        Article.objects.create(pub_date=now, headline='bar')
        Article.objects.create(pub_date=now, headline='AbBa')
        Article.objects.create(pub_date=now, headline='baz')
        Article.objects.create(pub_date=now, headline='baxZ')
        # zero-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'fo*'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'fo*'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
            ],
        )
        # one-or-more
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'fo+'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: hey-Foo>'
            ],
        )
        # wildcard
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'fooo?'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: hey-Foo>',
            ],
        )
        # leading anchor
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'^b'),
            [
                '<Article: AbBa>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'^a'),
            [
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        # trailing anchor
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'z$'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'z$'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>'
            ]
        )
        # character sets
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'ba[rz]'),
            [
                '<Article: AbBa>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>'
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'ba.[RxZ]'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'ba[RxZ]'),
            [
                '<Article: AbBa>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
            ],
        )

        # and more articles:
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')

        # alternation
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'oo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'oo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: zoocarfaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'^foo(f|b)'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: barfoobaz>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: bazbaRFOO>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',

            ],
        )

        # greedy matching
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'b.*az'),
            [
                '<Article: AbBa>',
                '<Article: bar>',
                '<Article: baxZ>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: foobar>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ]
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'b.*ar'),
            [
                '<Article: AbBa>',
                '<Article: baxZ>',
                '<Article: baz>',
                '<Article: f>',
                '<Article: fo>',
                '<Article: foo>',
                '<Article: foobaz>',
                '<Article: fooo>',
                '<Article: hey-Foo>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
            ]
        )

    @skipUnlessDBFeature('supports_exregex_backreferencing')
    def test_negate_exregex_backreferencing(self):
        """Copy from test_negate_regex_backreferencing."""
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobar')
        Article.objects.create(pub_date=now, headline='foobaz')
        Article.objects.create(pub_date=now, headline='ooF')
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='zoocarfaz')
        Article.objects.create(pub_date=now, headline='barfoobaz')
        Article.objects.create(pub_date=now, headline='bazbaRFOO')
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'b(.).*b\1'),
            [
                '<Article: foobar>',
                '<Article: foobaz>',
                '<Article: ooF>',
                '<Article: zoocarfaz>',
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )

    @skipIfDBVendor('postgresql', 'mysql')
    def test_negate_exregex_null_sqlite(self):
        """Copy from test_negate_regex_null_sqlite."""
        Season.objects.create(year=2012, gt=None)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neexregex=r'^$'),
            [
                '<Season: 2012>',
            ],
        )

    @skipIfDBVendor('sqlite')
    def test_negate_exregex_null_without_sqlite(self):
        """Copy from test_negate_regex_null_without_sqlite."""
        Season.objects.create(year=2012, gt=None)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neexregex=r'^$'),
            [],
        )

    def test_negate_exregex_non_string(self):
        """Copy from test_negate_regex_non_string."""
        Season.objects.create(year=2013, gt=444)
        self.assertQuerysetEqual(
            Season.objects.filter(gt__neexregex=r'^444$'),
            [],
        )

    def test_negate_exregex_non_ascii(self):
        """Copy from test_negate_regex_non_ascii."""
        Player.objects.create(name='\u2660')
        with self.assertRaises(Player.DoesNotExist):
            Player.objects.get(name__neexregex='\u2660')

    @skipIfDBVendor('sqlite')
    def test_exregex_word_starts_or_ends_brackets(self):
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='foo barbaz')
        Article.objects.create(pub_date=now, headline='Foo Barbaz')
        Article.objects.create(pub_date=now, headline='foobar baz')
        Article.objects.create(pub_date=now, headline='FooBar baz')
        Article.objects.create(pub_date=now, headline='foo bar baz')
        Article.objects.create(pub_date=now, headline='Foo Bar Baz')

        # [[:<:]] [[:>:]]
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'\<Bar\>'),
            [
                '<Article: Foo Bar Baz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exiregex=r'\<bar\>').order_by('id'),
            [
                '<Article: foo bar baz>',
                '<Article: Foo Bar Baz>',
            ],
        )

    @skipIfDBVendor('sqlite')
    def test_negate_exregex_word_starts_or_ends_brackets(self):
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobarbaz')
        Article.objects.create(pub_date=now, headline='foo barbaz')
        Article.objects.create(pub_date=now, headline='Foo Barbaz')
        Article.objects.create(pub_date=now, headline='foobar baz')
        Article.objects.create(pub_date=now, headline='FooBar baz')
        Article.objects.create(pub_date=now, headline='foo bar baz')
        Article.objects.create(pub_date=now, headline='Foo Bar Baz')

        # [[:<:]] [[:>:]]
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'\<Bar\>')
                           .exclude(headline__exregex=r'\<Article\>')
                           .order_by('id'),
            [
                '<Article: foobarbaz>',
                '<Article: foo barbaz>',
                '<Article: Foo Barbaz>',
                '<Article: foobar baz>',
                '<Article: FooBar baz>',
                '<Article: foo bar baz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexiregex=r'\<bar\>')
                           .exclude(headline__exregex=r'\<Article\>')
                           .order_by('id'),
            [
                '<Article: foobarbaz>',
                '<Article: foo barbaz>',
                '<Article: Foo Barbaz>',
                '<Article: foobar baz>',
                '<Article: FooBar baz>',
            ]
        )

    def test_exregex_short_end_escapes(self):
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobarbaz')

        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'\w+\s\d'),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'\W'),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'^\D+$'),
            [
                '<Article: foobarbaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__exregex=r'^\S+$'),
            [
                '<Article: foobarbaz>',
            ],
        )

    def test_negate_exregex_short_end_escapes(self):
        now = datetime.now()
        Article.objects.create(pub_date=now, headline='foobarbaz')

        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'\w+\s\d'),
            [
                '<Article: foobarbaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'\W'),
            [
                '<Article: foobarbaz>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'^\D+$'),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )
        self.assertQuerysetEqual(
            Article.objects.filter(headline__neexregex=r'^\S+$'),
            [
                '<Article: Article 5>',
                '<Article: Article 6>',
                '<Article: Article 4>',
                '<Article: Article 2>',
                '<Article: Article 3>',
                '<Article: Article 7>',
                '<Article: Article 1>',
            ],
        )

    def test_complement(self):
        tags = Tag.objects.filter(articles=OuterRef('id'), name='Tag 2')
        self.assertQuerysetEqual(
            Article.objects.filter(tag__complement=Exists(tags)).filter(author=self.au1),
            [
                '<Article: Article 4>',
                '<Article: Article 3>',
            ],
        )

    def test_negate_complement(self):
        tags = Tag.objects.filter(articles=OuterRef('id'), name='Tag 2')
        self.assertQuerysetEqual(
            Article.objects.filter(tag__complement=~Exists(tags)).filter(author=self.au1),
            [
                '<Article: Article 2>',
                '<Article: Article 1>',
            ],
        )



from django.db.backends.sqlite3.base import \
    DatabaseWrapper as DjangoDatabaseWrapper

from lookup_extensions.utils import merge_dicts

from .operations import DatabaseOperations


class ExtendedDatabaseWrapperMixin(object):
    ops_class = DatabaseOperations
    operators = merge_dicts(
        DjangoDatabaseWrapper.operators,
        {
            # For negates
            'neexact': '<> %s',
            'neiexact': "NOT LIKE %s ESCAPE '\\'",
            'necontains': "NOT LIKE %s ESCAPE '\\'",
            'neicontains': "NOT LIKE %s ESCAPE '\\'",
            'neregex': 'NOT REGEXP %s',
            'neiregex': "NOT REGEXP '(?i)' || %s",
            'nestartswith': "NOT LIKE %s ESCAPE '\\'",
            'neendswith': "NOT LIKE %s ESCAPE '\\'",
            'neistartswith': "NOT LIKE %s ESCAPE '\\'",
            'neiendswith': "NOT LIKE %s ESCAPE '\\'",
            # For exregex
            'exregex': 'REGEXP %s',
            'exiregex': "REGEXP '(?i)' || %s",
            'neexregex': 'NOT REGEXP %s',
            'neexiregex': "NOT REGEXP '(?i)' || %s",
        }
    )
    pattern_ops = merge_dicts(
        DjangoDatabaseWrapper.pattern_ops,
        {
            'necontains': r"NOT LIKE '%%' || {} || '%%' ESCAPE '\'",
            'neicontains': r"NOT LIKE '%%' || UPPER({}) || '%%' ESCAPE '\'",
            'nestartswith': r"NOT LIKE {} || '%%' ESCAPE '\'",
            'neistartswith': r"NOT LIKE UPPER({}) || '%%' ESCAPE '\'",
            'neendswith': r"NOT LIKE '%%' || {} ESCAPE '\'",
            'neiendswith': r"NOT LIKE '%%' || UPPER({}) ESCAPE '\'",
        }
    )
    regex_synonyms = {}


class DatabaseWrapper(ExtendedDatabaseWrapperMixin, DjangoDatabaseWrapper):
    pass



#!/usr/bin/env python

import os
import sys

import django

from .bootstrap import DJANGO_TEST_APP_PATH

os.environ.setdefault("DJANGO_SETTINGS_MODULE", 'test_settings')
sys.path.append(os.path.abspath(DJANGO_TEST_APP_PATH))
sys.path.append(os.path.abspath(os.path.dirname(__file__)))


def runtests():
    from django.conf import settings
    from django.test.utils import get_runner

    django.setup()

    from .monkey_patch import replace_managers

    replace_managers()

    TestRunner = get_runner(settings)
    test_runner = TestRunner(
        verbosity=2,
        interactive=False,
        keepdb=True,
        debug_sql=True,
    )
    if os.environ.get('TEST_WITH_REDSHIFT', None) == 'yes':
        test_runner.keepdb = True
    failures = test_runner.run_tests([
        'test_lookup',
        'test_or_lookups',
        'test_reverse_lookup',
        'test_string_lookup',
    ])
    sys.exit(bool(failures))


if __name__ == "__main__":
    runtests()



import environ

# Replace backends
environ.Env.DB_SCHEMES['postgres'] = 'lookup_extensions.backends.postgresql'
environ.Env.DB_SCHEMES['mysql'] = 'lookup_extensions.backends.mysql'
environ.Env.DB_SCHEMES['sqlite'] = 'lookup_extensions.backends.sqlite3'
env = environ.Env()


SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'lookup',
    'or_lookups',
    'reverse_lookup',
    'string_lookup',
]
DATABASES = {
    'default': env.db(default='sqlite://:memory:'),
}
if 'lookup_extensions.backends.mysql' == DATABASES['default']['ENGINE']:
    """Create database with specific options for MySQL.
    https://docs.djangoproject.com/en/dev/topics/testing/overview/#the-test-database
    https://docs.djangoproject.com/en/dev/ref/databases/#creating-your-database
    """
    DATABASES['default']['TEST'] = {
        'CHARSET': 'utf8mb4',
        'COLLATION': 'utf8mb4_unicode_ci',
    }



from django.db.backends.mysql.operations import \
    DatabaseOperations as DjangoDatabaseOperations

from lookup_extensions.backends.base.operations import (
    ExtendedDatabaseOperationsMixin,
)


class DatabaseOperations(ExtendedDatabaseOperationsMixin, DjangoDatabaseOperations):
    pass



from __future__ import unicode_literals

try:
    from lookup.test_lookups import YearComparisonLookupTests as DjangoYearComparisonLookupTests
except ImportError:
    # Django>=2.0 only
    from django.test import TestCase as DjangoYearComparisonLookupTests


class YearComparisonLookupTests(DjangoYearComparisonLookupTests):
    pass



from __future__ import unicode_literals

from operator import attrgetter
from unittest import skip

from django.db.models import Q

try:
    from or_lookups.models import Article
    from or_lookups.tests import OrLookupsTests as DjangoOrLookupsTests
except ImportError:
    print("Run tests/bootstrap.py before run tests.")
    raise


class OrLookupsTests(DjangoOrLookupsTests):
    def test_negate_filter_or(self):
        self.assertQuerysetEqual(
            (
                Article.objects.filter(headline__nestartswith='Hello') |
                Article.objects.filter(headline__nestartswith='Goodbye')
            ),
            [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline")
        )

        self.assertQuerysetEqual(
            Article.objects.filter(headline__necontains='Hello') | Article.objects.filter(headline__necontains='bye'),
            [
                'Hello',
                'Goodbye',
            ],
            attrgetter("headline")
        )

        self.assertQuerysetEqual(
            Article.objects.filter(headline__neiexact='Hello') | Article.objects.filter(headline__necontains='ood'),
            [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline")
        )

        self.assertQuerysetEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello') | Q(headline__nestartswith='Goodbye')),
            [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline")
        )

    def test_negate_stages(self):
        articles = Article.objects.all()
        self.assertQuerysetEqual(
            articles.filter(headline__nestartswith='Hello') & articles.filter(headline__nestartswith='Goodbye'),
            []
        )
        self.assertQuerysetEqual(
            articles.filter(headline__nestartswith='Hello') & articles.filter(headline__necontains='bye'),
            [],
            attrgetter("headline")
        )

    @skip("Support NOT IN filter")
    def test_negate_pk_in(self):
        self.assertQuerysetEqual(
            Article.objects.filter(pk__nein=[self.a1, self.a2, self.a3]), [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline"),
        )

        self.assertQuerysetEqual(
            Article.objects.filter(pk__nein=(self.a1, self.a2, self.a3)), [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline"),
        )

        self.assertQuerysetEqual(
            Article.objects.filter(pk__nein=[self.a1, self.a2, self.a3, 40000]), [
                'Hello',
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline"),
        )

    @skip("Support NOT IN filter")
    def test_negate_empty_in(self):
        # Passing "in" an empty list returns no results ...
        self.assertQuerysetEqual(
            Article.objects.filter(pk__nein=[]),
            []
        )
        # ... but can return results if we OR it with another query.
        self.assertQuerysetEqual(
            Article.objects.filter(Q(pk__nein=[]) | Q(headline__neicontains='goodbye')), [
                'Goodbye',
                'Hello and goodbye'
            ],
            attrgetter("headline"),
        )

    def test_negate_q_and(self):
        # Q arg objects are ANDed
        self.assertQuerysetEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello'), Q(headline__necontains='bye')),
            [],
            attrgetter("headline")
        )
        # Q arg AND order is irrelevant
        self.assertQuerysetEqual(
            Article.objects.filter(Q(headline__necontains='bye'), headline__nestartswith='Hello'),
            [],
            attrgetter("headline"),
        )

        self.assertQuerysetEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello') & Q(headline__nestartswith='Goodbye')),
            []
        )

    def test_negate_q_exclude(self):
        self.assertQuerysetEqual(
            Article.objects.exclude(Q(headline__nestartswith='Hello')),
            [
                'Hello',
                'Hello and goodbye',
            ],
            attrgetter("headline")
        )

    def test_negate_other_arg_queries(self):
        with self.assertRaises(Article.DoesNotExist):
            Article.objects.get(Q(headline__nestartswith='Hello'), Q(headline__necontains='bye'))

        self.assertEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello') | Q(headline__necontains='bye')).count(),
            2,
        )

        self.assertSequenceEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello'), Q(headline__necontains='bye')).values(),
            [],
        )

        self.assertEqual(
            Article.objects.filter(Q(headline__nestartswith='Hello')).in_bulk([self.a1, self.a2]),
            {self.a2: Article.objects.get(pk=self.a2)}
        )



from __future__ import unicode_literals

try:
    from lookup.test_decimalfield import DecimalFieldLookupTests as DjangoDecimalFieldLookupTests
except ImportError:
    print("Run tests/bootstrap.py before run tests.")
    raise


class DecimalFieldLookupTests(DjangoDecimalFieldLookupTests):
    pass



from __future__ import unicode_literals

from django.core.exceptions import FieldError

try:
    from reverse_lookup.models import Poll, User
    from reverse_lookup.tests import ReverseLookupTests as DjangoReverseLookupTests
except ImportError:
    print("Run tests/bootstrap.py before run tests.")
    raise


class ReverseLookupTests(DjangoReverseLookupTests):
    def test_negate_reverse_by_field(self):
        u1 = User.objects.get(
            poll__question__neexact="What's the first question?"
        )
        self.assertEqual(u1.name, "Jim Bo")

        u2 = User.objects.get(
            poll__question__neexact="What's the second question?"
        )
        self.assertEqual(u2.name, "John Doe")

    def test_negate_reverse_by_related_name(self):
        with self.assertRaises(Poll.DoesNotExist):
            Poll.objects.get(poll_choice__name__neexact="This is the answer.")

        p2 = Poll.objects.get(
            related_choice__name__neexact="This is not the answer.")
        self.assertEqual(p2.question, "What's the second question?")

    def test_negate_reverse_field_name_disallowed(self):
        msg = (
            "Cannot resolve keyword 'choice' into field. Choices are: "
            "creator, creator_id, id, poll_choice, question, related_choice"
        )
        with self.assertRaisesMessage(FieldError, msg):
            Poll.objects.get(choice__name__neexact="This is the answer")



from django.db.models import Exists
from django.db.models.fields import Field
from django.db.models.lookups import (
    BuiltinLookup,
    FieldGetDbPrepValueMixin,
)


@Field.register_lookup
class Complement(FieldGetDbPrepValueMixin, BuiltinLookup):
    lookup_name = 'complement'

    def process_rhs(self, compiler, connection):
        if self.rhs_is_direct_value() or not isinstance(self.rhs, Exists):
            raise ValueError("Exists subqueries are required")

        db_rhs = getattr(self.rhs, '_db', None)
        if db_rhs is not None and db_rhs != connection.alias:
            raise ValueError("Subqueries aren't allowed across different databases")

        return super(Complement, self).process_rhs(compiler, connection)

    def as_sql(self, compiler, connection):
        return self.process_rhs(compiler, connection)



from __future__ import unicode_literals

try:
    from lookup.test_timefield import TimeFieldLookupTests as DjangoTimeFieldLookupTests
except ImportError:
    print("Run tests/bootstrap.py before run tests.")


class TimeFieldLookupTests(DjangoTimeFieldLookupTests):
    pass



from django.db.models.constants import LOOKUP_SEP
from django.db.models.sql import Query as DjangoQuery


class ExtendedQueryMixin(object):
    SUBQUERY_LOOKUPS = (
        'complement',
    )

    def build_lookup(self, lookups, lhs, rhs):
        if lookups and lookups[-1] in self.SUBQUERY_LOOKUPS:
            return super(ExtendedQueryMixin, self).build_lookup(lookups[-1:], rhs, rhs)
        return super(ExtendedQueryMixin, self).build_lookup(lookups, lhs, rhs)

    def prepare_lookup_value(self, value, lookups, can_reuse, allow_joins=True):
        if lookups and lookups[-1] in self.SUBQUERY_LOOKUPS:
            value = value.resolve_expression(self, allow_joins=True, reuse=None, summarize=False)
            return value, lookups, []
        return super(ExtendedQueryMixin, self).prepare_lookup_value(
            value, lookups, can_reuse, allow_joins=allow_joins)

    def solve_lookup_type(self, lookup):
        lookup_splitted = lookup.split(LOOKUP_SEP)
        if lookup_splitted and lookup_splitted[-1] in self.SUBQUERY_LOOKUPS:
            return lookup_splitted, (), True  # No JOIN
        return super(ExtendedQueryMixin, self).solve_lookup_type(lookup)


class Query(ExtendedQueryMixin, DjangoQuery):
    pass



from django.db.backends.postgresql.base import \
    DatabaseWrapper as DjangoDatabaseWrapper

from lookup_extensions.utils import merge_dicts

from .operations import DatabaseOperations


class ExtendedDatabaseWrapperMixin(object):
    ops_class = DatabaseOperations
    operators = merge_dicts(
        DjangoDatabaseWrapper.operators,
        {
            # For negates
            'neexact': '<> %s',
            'neiexact': '<> UPPER(%s)',
            'necontains': 'NOT LIKE %s',
            'neicontains': 'NOT LIKE UPPER(%s)',
            'neregex': '!~ %s',
            'neiregex': '!~* %s',
            'nestartswith': 'NOT LIKE %s',
            'neendswith': 'NOT LIKE %s',
            'neistartswith': 'NOT LIKE UPPER(%s)',
            'neiendswith': 'NOT LIKE UPPER(%s)',
            # For exregex
            'exregex': '~ %s',
            'exiregex': '~* %s',
            'neexregex': '!~ %s',
            'neexiregex': '!~* %s',
        }
    )
    pattern_ops = merge_dicts(
        DjangoDatabaseWrapper.pattern_ops,
        {
           'necontains': r"NOT LIKE '%%' || {} || '%%'",
           'neicontains': r"NOT LIKE '%%' || UPPER({}) || '%%'",
           'nestartswith': r"NOT LIKE {} || '%%'",
           'neistartswith': r"NOT LIKE UPPER({}) || '%%'",
           'neendswith': r"NOT LIKE '%%' || {}",
           'neiendswith': r"NOT LIKE '%%' || UPPER({})",
        }
    )
    regex_synonyms = {
        '\\<': '[[:<:]]',
        '\\>': '[[:>:]]',
    }


class DatabaseWrapper(ExtendedDatabaseWrapperMixin, DjangoDatabaseWrapper):
    pass



from __future__ import unicode_literals

try:
    from string_lookup.models import Article, Foo
    from string_lookup.tests import StringLookupTests as DjangoStringLookupTests
except ImportError:
    raise


class StringLookupTests(DjangoStringLookupTests):
    def test_negate_unicode_chars_in_queries(self):
        fx = Foo(name='Bjorn', friend='CHANGED')
        fx.save()
        self.assertEqual(Foo.objects.get(friend__necontains='\xe7'), fx)

    def test_negate_queries_on_textfields(self):
        a = Article(name='Test', text='CHANGED')
        a.save()
        self.assertEqual(Article.objects.get(text__neexact='The quick brown fox jumps over the lazy dog.'), a)

        self.assertEqual(Article.objects.get(text__necontains='quick brown fox'), a)

    def test_negate_ipaddress_on_postgresql(self):
        a = Article(name='IP test', text='The body', submitted_from='192.0.1.100')  # CHANGED
        a.save()
        self.assertSequenceEqual(Article.objects.filter(submitted_from__necontains='192.0.2'), [a])
        self.assertEqual(Article.objects.filter(submitted_from__necontains='32').count(), 1)



#!/usr/bin/env python

import logging
import os

import django
from django.utils.six.moves.urllib import request
from django.utils.six.moves.urllib.error import HTTPError

logger = logging.getLogger()


DJANGO_VERSION = django.get_version()
if 'dev' in DJANGO_VERSION:
    DJANGO_VERSION = 'master'
DJANGO_TEST_APP_PATH = os.path.join(
    os.path.dirname(os.path.dirname(__file__)),
    'vendor',
    'django',
    DJANGO_VERSION,
    'tests',
)
DJANGO_TEST_DOWNLOAD_URL = 'https://raw.githubusercontent.com/django/django/{version}/tests/{test_app}/{test_file}'
DJANGO_LOOKUP_TEST_APP_FILES = {
    'lookup': [
        'test_decimalfield.py',
        'test_lookups.py',
        'test_timefield.py',
    ],
    'or_lookups': [],
    'reverse_lookup': [],
    'string_lookup': [],
}


def download_django_test_apps():
    def _download_file(test_app, test_file):
        download_url = DJANGO_TEST_DOWNLOAD_URL.format(
            version=DJANGO_VERSION,
            test_app=test_app,
            test_file=test_file,
        )
        to_dir = os.path.join(
            DJANGO_TEST_APP_PATH,
            test_app,
        )
        if not os.path.isdir(to_dir):
            os.makedirs(to_dir)
        to_file = os.path.join(
            to_dir,
            test_file,
        )
        if not os.path.isfile(to_file):
            try:
                with open(to_file, 'wb') as fp:
                    res = request.urlopen(download_url)
                    fp.write(res.read())
                    logger.info("Download: %s", download_url)
                    res.close()
            except HTTPError:
                logger.warn("Not found: %s", download_url)

    for test_app, test_files in DJANGO_LOOKUP_TEST_APP_FILES.items():
        _download_file(test_app, '__init__.py')
        _download_file(test_app, 'models.py')
        _download_file(test_app, 'tests.py')
        for test_file in test_files:
            _download_file(test_app, test_file)


if __name__ == "__main__":
    download_django_test_apps()



# Load lookup types
from lookup_extensions import lookups  # noqa F401


class ExtendedDatabaseOperationsMixin(object):
    NEGATE_REGULAR_LOOKUP_TYPES = {
        'neexact': 'exact',
        'neiexact': 'iexact',
        'necontains': 'contains',
        'neicontains': 'icontains',
        'neregex': 'regex',
        'neiregex': 'iregex',
        'exregex': 'regex',
        'exiregex': 'iregex',
        'neexregex': 'regex',
        'neexiregex': 'iregex',
        'nestartswith': 'startswith',
        'neistartswith': 'istartswith',
        'neendswith': 'endswith',
        'neiendswith': 'iendswith',
    }

    def convert_to_regular_lookup_type(self, lookup_type):
        return self.NEGATE_REGULAR_LOOKUP_TYPES.get(lookup_type, None)

    def lookup_cast(self, lookup_type, internal_type=None):
        regular_lookup_type = self.convert_to_regular_lookup_type(lookup_type)
        if regular_lookup_type:
            return super(ExtendedDatabaseOperationsMixin, self).lookup_cast(
                regular_lookup_type, internal_type=internal_type)
        return super(ExtendedDatabaseOperationsMixin, self).lookup_cast(
            lookup_type, internal_type=internal_type)



from django.db.backends.mysql.base import \
    DatabaseWrapper as DjangoDatabaseWrapper

from lookup_extensions.utils import merge_dicts

from .operations import DatabaseOperations


class ExtendedDatabaseWrapperMixin(object):
    ops_class = DatabaseOperations
    operators = merge_dicts(
        DjangoDatabaseWrapper.operators,
        {
            # For negates
            'neexact': '<> %s',
            'neiexact': 'NOT LIKE %s',
            'necontains': 'NOT LIKE BINARY %s',
            'neicontains': 'NOT LIKE %s',
            'neregex': 'NOT REGEXP BINARY %s',
            'neiregex': 'NOT REGEXP %s',
            'nestartswith': 'NOT LIKE BINARY %s',
            'neendswith': 'NOT LIKE BINARY %s',
            'neistartswith': 'NOT LIKE %s',
            'neiendswith': 'NOT LIKE %s',
            # For exregex
            'exregex': 'REGEXP BINARY %s',
            'exiregex': 'REGEXP %s',
            'neexregex': 'NOT REGEXP BINARY %s',
            'neexiregex': 'NOT REGEXP %s',
        }
    )
    pattern_ops = merge_dicts(
        DjangoDatabaseWrapper.pattern_ops,
        {
            'necontains': "NOT LIKE BINARY CONCAT('%%', {}, '%%')",
            'neicontains': "NOT LIKE CONCAT('%%', {}, '%%')",
            'nestartswith': "NOT LIKE BINARY CONCAT({}, '%%')",
            'neistartswith': "NOT LIKE CONCAT({}, '%%')",
            'neendswith': "NOT LIKE BINARY CONCAT('%%', {})",
            'neiendswith': "NOT LIKE CONCAT('%%', {})",
        }
    )
    regex_synonyms = {
        '\\<': '[[:<:]]',
        '\\>': '[[:>:]]',
        '\\d': '[[:digit:]]',
        '\\D': '[^[:digit:]]',
        '\\w': '[[:alnum:]_]',
        '\\W': '[^[:alnum:]_]',
        '\\s': '[[:space:]]',
        '\\S': '[^[:space:]]',
    }


class DatabaseWrapper(ExtendedDatabaseWrapperMixin, DjangoDatabaseWrapper):
    pass
